/*
 * segment.c - NILFS segment constructor.
 *
 * Copyright (C) 2005 Nippon Telegraph and Telephone Corporation.
 *
 * This file is part of NILFS.
 *
 * NILFS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NILFS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NILFS; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * segment.c,v 1.155 2006/01/26 07:15:04 ryusuke Exp
 *
 * Written by Ryusuke Konishi <ryusuke@osrg.net>
 *
 */

#include <linux/pagemap.h>
#include <linux/buffer_head.h>
#include <linux/writeback.h>
#include <linux/crc32c.h>
#include <linux/bio.h>
#include <linux/completion.h>
#include <linux/blkdev.h>
#include <linux/suspend.h>
#include "nilfs.h"
#include "sb.h"

static void nilfs_segctor_commit(struct nilfs_sc_info *sci);

#define nilfs_cnt32_gt(a,b)   \
        (typecheck(__u32, a) && typecheck(__u32, b) && \
         ((__s32)(b) - (__s32)(a) < 0))
#define nilfs_cnt32_ge(a,b)   \
        (typecheck(__u32, a) && typecheck(__u32, b) && \
         ((__s32)(a) - (__s32)(b) >= 0))
#define nilfs_cnt32_lt(a,b)   nilfs_cnt32_gt(b,a)
#define nilfs_cnt32_le(a,b)   nilfs_cnt32_ge(b,a)

/* 
 * Transaction 
 *
 * We don't need the exclusion control among same task, because
 * all file operations are serialized through inode->i_sem including 
 * ones by the same task.
 */
static kmem_cache_t *nilfs_transaction_cachep;

/**
 * nilfs_init_transaction_cache - create a cache for nilfs_transaction_info
 *
 * nilfs_init_transaction_cache() creates a slab cache for the struct
 * nilfs_transaction_info. 
 *
 * Return Value: On success, it returns 0. On error, a following negative
 * error code is returned.
 * 
 * %-ENOMEM - Insufficient memory available.
 */
int nilfs_init_transaction_cache(void)
{
	nilfs_transaction_cachep =
		kmem_cache_create("nilfs_transaction_cache",
				  sizeof(struct nilfs_transaction_info),
				  0, SLAB_RECLAIM_ACCOUNT,
				  NULL, NULL);
  
	return ((nilfs_transaction_cachep == NULL) ? -ENOMEM : 0);
}

/**
 * nilfs_detroy_transaction_cache - destroy the cache for nilfs_transaction_info
 *
 * nilfs_destroy_transaction_cache() frees the slab cache for the struct
 * nilfs_transaction_info.
 */
void nilfs_destroy_transaction_cache(void)
{
	if (kmem_cache_destroy(nilfs_transaction_cachep))
		printk(KERN_INFO
		       "NILFS: some transaction on memory were not freed\n");
}

/**
 * nilfs_prepare_segment_lock - prepare to lock the segment semaphore.
 * @sb: super block
 * @ti: nilfs_transaction_info
 *
 * nilfs_prepare_segment_lock() sets up or updates the struct 
 * nilfs_transaction_info into the journal_info field of task_struct of
 * the current task. If the current task holds a valid 
 * nilfs_transaction_info, its ti_count field is incremented.
 * If the current task doesn't have a nilfs_transaction_info,
 * given nilfs_transaction_info by @ti or (when @ti is NULL) dynamically
 * allocated one is initialized and is hooked on journal_info of
 * the current task.
 * In the later case, %NILFS_TI_DYNAMIC_ALLOC flag is set on ti_flags.
 *
 * Return Value: On success, the value of ti_count (nest level count) is
 * returned. On error, a following negative error code is returned.
 * 
 * %-ENOMEM - Insufficient memory available.
 */
static int
nilfs_prepare_segment_lock(struct super_block *sb, struct nilfs_transaction_info *ti)
{
	struct nilfs_transaction_info *cur_ti = current->journal_info;
	void *save = NULL;

	if (cur_ti) {
		if (cur_ti->ti_super == sb) {
			return ++cur_ti->ti_count;
		} else {
			/*
			 * If journal_info field is occupied by other filesystem,
			 * we save it and restore on nilfs_transaction_end().
			 * But this should never happen.
			 */
			nilfs_warning(sb, __FUNCTION__, 
				      "journal info from a different FS");
			save = current->journal_info;
		}
	}
	if (!ti) {
		ti = kmem_cache_alloc(nilfs_transaction_cachep, SLAB_NOFS);
		if (!ti)
			return -ENOMEM;
		ti->ti_flags = NILFS_TI_DYNAMIC_ALLOC;
	} else {
		ti->ti_flags = 0;
	}
	ti->ti_count = 0;
	ti->ti_save = save;
	ti->ti_super = sb;
	current->journal_info = ti;
	return 0;
}

/**
 * nilfs_transaction_begin - start indivisible file operations.
 * @sb: super block
 * @ti: nilfs_transaction_info
 *
 * nilfs_transaction_begin() acquires the read semaphore that excludes
 * segment construction as needed. This function is used with 
 * nilfs_transaction_end() in pairs. The region enclosed by these
 * functions excludes the segment construction.  Their enclosure can be
 * nested; acquiring and releasing of the read semaphore is performed
 * only in the outermost call to avoid multiple read locking by the same
 * task.
 * @ti is used to specify a nilfs_transaction_info on local memory.
 * In outermost call, the memory is initialized and is hooked onto the 
 * current task. The allocation of nilfs_transaction_info can be 
 * done dynamically by giving NULL to @ti.
 * 
 * Return Value: On success, 0 is returned. On error, a following negative
 * error code is returned.
 *
 * %-ENOMEM - Insufficient memory available.
 */
int nilfs_transaction_begin(struct super_block *sb, struct nilfs_transaction_info *ti)
{
	int ret = nilfs_prepare_segment_lock(sb, ti);

	if (ret < 0)
		return ret;
	if (ret == 0) {
		struct nilfs_sb_info *sbi;

		might_sleep();
		seg_debug(3, "task %p locking segment semaphore\n", current);
		down_read(&NILFS_SB(sb)->s_segctor.sem);
		seg_debug(3, "locked\n");
		sbi = NILFS_SB(sb);
		if (NILFS_SC(sbi)) 
			nilfs_segctor_commit(NILFS_SC(sbi));
	} else
		seg_debug(3, "incremented transaction refcnt (ti=%p, cnt=%d)\n",
			  current->journal_info, 
			  ((struct nilfs_transaction_info *)current->journal_info)->ti_count);
	return 0;
}

/**
 * nilfs_transaction_end - end indivisible file operations.
 * @sb: super block
 *
 * nilfs_transaction_end() releases the read semaphore which is 
 * acquired by nilfs_transaction_begin(). Its releasing is only done
 * in outermost call of this function. If the nilfs_transaction_info
 * was allocated dynamically, it is given back to a slab cache.
 */
int nilfs_transaction_end(struct super_block *sb)
{
	struct nilfs_transaction_info *ti = current->journal_info;
	struct nilfs_sb_info *sbi;
	int err = 0;

	if (ti == NULL || ti->ti_super != sb) {
		seg_debug(1, "missing nilfs_transaction_begin()\n");
		BUG();
	}
	if (ti->ti_count > 0) {
		ti->ti_count--;
		seg_debug(3, "decremented transaction refcnt (ti=%p, cnt=%d)\n",
			  ti, ti->ti_count);
		return 0;
	}
	sbi = NILFS_SB(sb);
	up_read(&sbi->s_segctor.sem);
	seg_debug(3, "task %p unlocked segment semaphore\n", current);
	current->journal_info = ti->ti_save;
	if (ti->ti_flags & NILFS_TI_SYNC) {
		struct writeback_control wbc = {
			.sync_mode = WB_SYNC_ALL,
			.nr_to_write = 0,
			.nonblocking = 0,
		};
		err = nilfs_construct_segment(sb, &wbc);
		if (!NILFS_SEG_ERR(err))
			err = 0;
	}
	if (ti->ti_flags & NILFS_TI_DYNAMIC_ALLOC)
		kmem_cache_free(nilfs_transaction_cachep, ti);
	return err;
}

/* 
 * Segment constructor
 */
#define SC_N_PAGEVEC      16    /* size of locally allocated page vector */
#define SC_N_BHVEC        16    /* size of locally allocated buffer head
				   vector */

#define SC_BYTES_SECTOR(b)      ((((b) - 1) >> NILFS_SECTOR_BITS) + 1)


/* Internal return codes */
#define NILFS_SEG_FULL    1

/* Collection stages */
enum {
	SC_COLLECT_INIT = 0,
	SC_COLLECT_DATA,
	SC_COLLECT_FBT,
	SC_COLLECT_IBLK,
	SC_COLLECT_IBT,
	SC_COLLECT_CP,
	SC_COLLECT_DONE,
};


#define SC_BH(sci,idx)   ((sci)->sc_bh_arr[idx])
#define SC_SUM_BH(sci,idx)  SC_BH(sci, (sci)->sc_ba_max - idx - 1)


typedef int (*data_proc_t)(struct buffer_head *, void *, fbn_t);
typedef int (*node_proc_t)(struct buffer_head *, void *);

static int lookup_dirty_buffers(struct super_block *, struct address_space *,
				void *, data_proc_t);
static int lookup_buffers_64(struct radix_tree_64_root *, spinlock_t *,
			     void *, node_proc_t, int);
static void abort_segment_io(struct nilfs_sc_info *, int);
static void complete_segment_io(struct nilfs_sc_info *);

/**
 * nilfs_segment_update_super - change pointer to the latest segment
 * @sb: super block
 *
 * nilfs_segment_update_super() changes information in the super block
 * after the logical segment is written out successfully. The super
 * block is marked dirty. It will be written out the next VFS sync
 * operations such as sync_supers() and generic_shutdown_super().
 */
static void nilfs_segment_update_super(struct super_block *sb)
{
	struct nilfs_sb_info *sbi = NILFS_SB(sb);
	struct the_nilfs *nilfs = sbi->s_nilfs;

	seg_debug(2, "detected change of last checkpoint\n");
	down_write(&nilfs->ns_sem);
	lock_super(sb);
	nilfs->ns_sbp->s_last_segment = cpu_to_le64(nilfs->ns_last_pseg);
	nilfs->ns_sbp->s_last_seq = cpu_to_le64(nilfs->ns_seg_seq);
	sb->s_dirt = 1;
	nilfs_mark_buffer_dirty(nilfs->ns_sbh);
	unlock_super(sb);
	up_write(&nilfs->ns_sem);
}

/**
 * wait_on_segment - wait for completion of requested BIOs
 * @sbi: nilfs_sb_info
 * @sci: nilfs_sc_info
 *
 * wait_on_segment() waits for the completion of currently requesting
 * BIOs and gets the result of submissions. When the construction of a
 * logical segment is completed successfully, it updates a pointer to the
 * latest partial segment by calling nilfs_segment_update_super().
 *
 * Return Value: On Success, 0 is returned. On Error, a following negative
 * error code is returned.
 *
 * %-EIO - I/O error
 */
static inline int
wait_on_segment(struct nilfs_sb_info *sbi, struct nilfs_sc_info *sci)
{
	int err = 0;

	seg_debug(2, "called nbio=%d\n", sci->sc_nbio);
	if (!sci->sc_nbio)
		return 0;

	do {
		wait_for_completion(&sci->sc_bio_event);
	} while (--sci->sc_nbio > 0);

	seg_debug(2, "wait completed\n");
	if (test_and_clear_bit(NILFS_SC_EIO, &sbi->s_segctor.flags)) {
		seg_debug(2, "found bio error\n");
		err = -EIO;
	}
	if (test_and_clear_bit(NILFS_SC_LAST_CP, &sbi->s_segctor.flags))
		nilfs_segment_update_super(sci->sc_sb);
	return err;
}

#ifdef CONFIG_NILFS_DEBUG
static void
print_segment_info(struct nilfs_sc_info *sci, dbn_t seg_start, dbn_t seg_end)
{
	struct the_nilfs *nilfs = NILFS_SB(sci->sc_sb)->s_nilfs;

	if (nilfs_debug_info.verbose[NILFS_VERBOSE_SEGINFO] == 1) {
		printk(KERN_DEBUG "NILFS(seginfo) pseg(blk=%lu,len=%u)"
		       " on fullseg(ind=%lu,blk=%lu->%lu)\n",
		       sci->sc_pseg_start, sci->sc_sum.nblocks,
		       nilfs->ns_last_segnum, seg_start, seg_end);
		return;
	}
	printk(KERN_DEBUG "========= NILFS SEGMENT INFORMATION ========\n");
	printk(KERN_DEBUG "full segment: segnum=%lu, start=%lu, end=%lu\n",
	       nilfs->ns_last_segnum, seg_start, seg_end);
	printk(KERN_DEBUG "partial segment: start=%lu, rest=%lu\n",
	       sci->sc_pseg_start, sci->sc_residual_blocks);
	printk(KERN_DEBUG "------------------ SUMMARY -----------------\n");
	printk(KERN_DEBUG "nfinfo     = %u (number of file)\n",
	       sci->sc_sum.nfinfo);
	printk(KERN_DEBUG "nfbinfo    = %u (number of fbinfo)\n",
	       sci->sc_sum.nfbinfo);
	printk(KERN_DEBUG "nblocks    = %u (number of blocks)\n",
	       sci->sc_sum.nblocks);
	printk(KERN_DEBUG "nblk_sum   = %u (number of summary blocks)\n",
	       sci->sc_sum.nblk_sum);
	printk(KERN_DEBUG "sum_bytes  = %lu (byte counts of summary)\n",
	       sci->sc_sum_bytes);
	printk(KERN_DEBUG "nblk_file  = %u (number of file blocks)\n",
	       sci->sc_sum.nblk_file);
	printk(KERN_DEBUG "nblk_fbt   = %u (number of file B-tree blocks)\n",
	       sci->sc_sum.nblk_fbt);
	printk(KERN_DEBUG "nblk_inode = %u (number of inode blocks)\n",
	       sci->sc_sum.nblk_inode);
	printk(KERN_DEBUG "nblk_ibt   = %u (number of inode B-tree blocks)\n",
	       sci->sc_sum.nblk_ibt);
	printk(KERN_DEBUG "nblk_cp    = %u (number of checkpoint blocks)\n",
	       NILFS_SEG_HAS_CP(&sci->sc_sum) ? 1 : 0);
	printk(KERN_DEBUG "============================================\n");
}
#endif

/**
 * nilfs_commit_dirty_file - register a file to the dirty file list
 * @inode: inode of the file to be registered.
 *
 * nilfs_commit_dirty_file() registers a file specified by @inode to
 * the dirty file list. This list is used to limit the scope of searches
 * of dirty blocks in the following segment construction.
 * If nilfs_inode of the file has a %NILFS_STATE_UPDATE flag on its i_state
 * field, it is reentered to the dirty file list and the 
 * %NILFS_STATE_UPDATE flag is cleared. The %NILFS_STATE_UPDATE flag 
 * indicates that the nilfs_inode has been removed from the dirty file list
 * but still has a reference of inode.
 * If the nilfs_inode does not belong to any list, it is simply added to
 * the dirty file list. If the registration succeeded, 
 * the %NILFS_STATE_DIRTY flag is set, the inode block which contains
 * the nilfs_inode is loaded and is marked dirty.
 * In any case, it accompanies a call of nilfs_transaction_end().
 *
 * Return Value: On success, 0 is returned. On error, a following negative
 * error code is returned.
 *
 * %-EINVAL - cannot grab the inode (This may happen when somebody is 
 * freeing the inode) or specified inode has an invalid inode number.
 *
 * %-EIO - I/O error
 */
int nilfs_commit_dirty_file(struct inode *inode)
{
	struct nilfs_sb_info *sbi;
	struct nilfs_inode_info *ii;
	struct nilfs_inode *raw_inode;
	struct buffer_head *ibh;
	int err, err2;

	BUG_ON(is_bad_inode(inode));
	sbi = NILFS_SB(inode->i_sb);
	ii = NILFS_I(inode);

#ifdef CONFIG_NILFS_DEBUG
	if (inode->i_state & I_FREEING)
		nilfs_warning(inode->i_sb, __FUNCTION__,
			      "trying to mark deleting file dirty.\n");
#endif	
	/* Readers need spinlock for accessing s_segctor.dirty_files */
	spin_lock_irq(&sbi->s_segctor.dirty_files_lock);
	raw_inode = __nilfs_load_inode_block_nolock(sbi, inode, &ibh);
	if (IS_ERR(raw_inode)) {
		/* We first try to get a buffer of the specified inode to
		   ensure that every inode in the dirty_files list has a
		   valid BH pointer in i_bh. */
		nilfs_warning(inode->i_sb, __FUNCTION__,
			      "failed to reget inode block.\n");
		err = PTR_ERR(raw_inode);
		spin_unlock_irq(&sbi->s_segctor.dirty_files_lock);
		goto fail;
	}

	if (ii->i_state & NILFS_STATE_UPDATED) { /* redirty file */
		list_del_init(&ii->i_dirty);
		ii->i_state &= ~NILFS_STATE_UPDATED;
		list_add_tail(&ii->i_dirty, &sbi->s_segctor.dirty_files);
		seg_debug(2, "redirtied file (ino=%lu)\n", inode->i_ino);
	} else if (list_empty(&ii->i_dirty)) {
		if (igrab(inode) == NULL) {
			/* This will happen when somebody is freeing 
			   this inode. */
			nilfs_warning(inode->i_sb, __FUNCTION__,
				      "cannot get inode (ino=%lu)\n",
				      inode->i_ino);
			spin_unlock_irq(&sbi->s_segctor.dirty_files_lock);
			brelse(ibh);
			err = -EINVAL;
			goto fail;
		}
		list_add_tail(&ii->i_dirty, &sbi->s_segctor.dirty_files);
		seg_debug(2, "registered dirty file (ino=%lu)\n", inode->i_ino);
	}
	ii->i_state |= NILFS_STATE_DIRTY;
	spin_unlock_irq(&sbi->s_segctor.dirty_files_lock);
	nilfs_mark_inode_buffer_dirty(sbi, ibh);
	brelse(ibh);
	err = 0;
 fail:
	err2 = nilfs_transaction_end(inode->i_sb);
	return (err ? : err2);
}

/**
 * nilfs_dirty_inode - reflect changes on given inode to an inode block.
 * @inode: inode of the file to be registered.
 *
 * nilfs_dirty_inode() loads a inode block containing the specified
 * @inode and copies data from a nilfs_inode to a corresponding inode
 * entry in the inode block. This operation is excluded from the segment
 * construction. This function can be called both as a single operation
 * and as a part of indivisible file operations.
 */
void nilfs_dirty_inode(struct inode *inode)
{
	struct nilfs_sb_info *sbi;
	struct nilfs_inode *raw_inode;
	struct buffer_head *ibh;
	struct nilfs_transaction_info ti;

	BUG_ON(is_bad_inode(inode));

	sbi = NILFS_SB(inode->i_sb);
	nilfs_transaction_begin(inode->i_sb, &ti);

	raw_inode = nilfs_load_inode_block(sbi, inode, &ibh);
	if (IS_ERR(raw_inode)) {
		nilfs_warning(inode->i_sb, __FUNCTION__,
			      "failed to reget inode block.\n");
		nilfs_transaction_end(inode->i_sb); /* never fails */
		return;
	}
	lock_buffer(ibh);
	nilfs_update_inode(inode, raw_inode, ibh);
	unlock_buffer(ibh);
	nilfs_mark_inode_buffer_dirty(sbi, ibh);
	brelse(ibh);
	nilfs_transaction_end(inode->i_sb); /* never fails */
}

/*
 * BIO operations
 */

/**
 * clear_seg_bio - clear cancelled bio
 * @sbi: nilfs_sb_info
 * @bio: cancelled BIO
 *
 * clear_seg_bio() releases buffer heads refered by the BIO page vector.
 */
static void
clear_seg_bio(struct nilfs_sb_info *sbi, struct bio *bio)
{
	struct bio_vec *bvec = bio->bi_io_vec + bio->bi_vcnt - 1;

	do {
		struct bio_vec *bv2 = bvec;
		struct buffer_head *bh, *head;
		unsigned int offset = 0;

		if (--bvec >= bio->bi_io_vec)
			prefetchw(&bvec->bv_page->flags);

		bh = head = page_buffers(bv2->bv_page);
		do {
			if (offset >= bv2->bv_offset) {
				if (offset >= bv2->bv_offset + bv2->bv_len)
					break;
				/* clear_buffer_uptodate(bh); */
				nilfs_brelse_isr(sbi->s_super, bh);
			}
			offset += bh->b_size;
			bh = bh->b_this_page;
		} while (bh != head);
	} while (bvec >= bio->bi_io_vec);
}

/**
 * end_seg_bio_write - complete a BIO for the segment write
 * @bio: cancelled BIO
 * @bytes_done: byte count of completed data
 * @err: err code notified by the BIO layer
 *
 * end_seg_bio_write() is a completion routine of BIO for segment write.
 * This routine performs end processing for the BIO.
 * When the block count reaches the number of block in this segment, it 
 * calls complete_segment_io() to finish a paritial segment. If an error
 * flag is set, it calles instead abort_segment_io() to cancel BIOs.
 *
 * Return Value: If given BIO has not finished yet, 1 is returned.
 * Otherwise, 0 is returned.
 */
static int
end_seg_bio_write(struct bio *bio, unsigned int bytes_done, int err)
{
	const int uptodate = test_bit(BIO_UPTODATE, &bio->bi_flags);
	struct nilfs_sb_info *sbi;
	struct nilfs_sc_info *sci;
	unsigned int blk_done;

	if (bio->bi_size)
		return 1;

	sbi = (struct nilfs_sb_info *)bio->bi_private;
	clear_seg_bio(sbi, bio);

	if (err == -EOPNOTSUPP) {
		struct bio_vec *bv, *ep = bio->bi_io_vec + bio->bi_vcnt - 1;

		set_bit(BIO_EOPNOTSUPP, &bio->bi_flags);
		for (bv = bio->bi_io_vec; bv <= ep; bv++)
			unlock_page(bv->bv_page);

		bio_put(bio);
		return 0; /* to be detected by submit_seg_bio() */
	}

	sci = NILFS_SC(sbi);
	blk_done = bytes_done >> sbi->s_super->s_blocksize_bits;

	if (!uptodate)
		set_bit(NILFS_SC_EIO, &sbi->s_segctor.flags);

	seg_debug(2, "blocks written by this bio = %u, "
		  "residual blocks of this segment = %u\n",
		  blk_done, atomic_read(&sci->sc_bio_blk_cnt));
	if (atomic_sub_and_test(blk_done, &sci->sc_bio_blk_cnt)) {
		if (test_bit(NILFS_SC_EIO, &sbi->s_segctor.flags)) {
			abort_segment_io(sci, sci->sc_sum.nblocks);
			seg_debug(1, "failed to writeout a segment\n");
		} else {
			complete_segment_io(sci);
			seg_debug(1, "completed a segment\n");
		}
	}

	bio_put(bio);
	complete(&sci->sc_bio_event);
	return 0;
}

/**
 * submit_seg_bio - submit a given bio of the segment.
 * @bio: bio
 * @sbi: nilfs_sb_info
 * @mode: write mode
 *
 * submit_seg_bio() sends a write request to the BIO layer.
 * The request is specified by @bio.
 * This function always locks pages and set their state writeback even
 * when it returns an error. If @sync is specified non-zero value,
 * bio is submitted with a WRITE_SYNC flag instead of a WRITE flag.
 *
 * Return Value: On success, 0 is returned. on error, a following negative
 * error code is returned.
 * 
 * %-EIO - I/O error.
 * %-EOPNOTSUPP - Operation is not supported.
 */
static int submit_seg_bio(struct bio *bio, struct nilfs_sb_info *sbi, int mode)
{
	struct nilfs_sc_info *sci = NILFS_SC(sbi);
	struct bio_vec *bv, *ep = bio->bi_io_vec + bio->bi_vcnt - 1;
	int err = 0;
	int max_bio = atomic_read(&sci->sc_max_bio);

	for (bv = bio->bi_io_vec; bv <= ep; bv++)
		lock_page(bv->bv_page);

	while (sci->sc_nbio >= max_bio) {
		seg_debug(2, "waiting for a segment\n");
		wait_for_completion(&sci->sc_bio_event);
		sci->sc_nbio--;
		if (test_bit(NILFS_SC_EIO, &sbi->s_segctor.flags)) {
			seg_debug(1, "detected error of a submitted segment\n");
			clear_seg_bio(sbi, bio);
			bio_put(bio);
			return -EIO;
		}
	}
#if 0  /* issue i/o-request immediately */
	else if (sci->sc_nbio + 1 == max_bio)
		mode |= (1 << BIO_RW_SYNC);
#endif

	bio->bi_end_io = end_seg_bio_write;
	bio->bi_private = sbi;
	seg_debug(2, "submitting bio (start_dbn=%llu, size=%u, vcnt=%hu, "
		  "barrier=%d)\n",
		  bio->bi_sector, bio->bi_size, bio->bi_vcnt,
		  (mode & (1 << BIO_RW_BARRIER)) != 0);

	bio_get(bio);
	submit_bio(mode, bio);
	if (bio_flagged(bio, BIO_EOPNOTSUPP)) {
		err = -EOPNOTSUPP;
		seg_debug(1, "aborted bio submission\n");
	}
	bio_put(bio);
	return err;
}

/** 
 * alloc_seg_bio - allocate a bio for writing segment.
 * @sb: super block
 * @start: beginning disk block number of this BIO.
 * @nr_vecs: request size of page vector.
 *
 * alloc_seg_bio() allocates a new BIO structure and initialize it.
 *
 * Return Value: On success, pointer to the struct bio is returned.
 * On error, NULL is returned.
 */
static struct bio *
alloc_seg_bio(struct super_block *sb, dbn_t start, int nr_vecs)
{
	struct bio *bio;

	bio = bio_alloc(GFP_NOIO, nr_vecs);
	if (bio == NULL && (current->flags & PF_MEMALLOC)) {
		while (!bio && (nr_vecs >>= 1))
			bio = bio_alloc(GFP_NOIO, nr_vecs);
	}
	if (bio) {
		bio->bi_bdev = sb->s_bdev;
		bio->bi_sector = (sector_t)start << (sb->s_blocksize_bits - NILFS_SECTOR_BITS);
		seg_debug(2, "allocated bio (max_vecs=%d)\n", bio->bi_max_vecs);
	}
	return bio;
}

/** 
 * seg_bio_add_bh - appends a buffer head to specified bio.
 * @bio: bio 
 * @bh: buffer head to be appended to the bio.
 *
 * seg_bio_add_bh() adds a buffer head @bh to a page vector of @bio.
 * It also gets the reference of the buffer and locks it.
 * When returning an error, it will do nothing to the buffer.
 *
 * Return Value: On success, 0 is returned. If the page vector is full,
 * 1 is returned.
 */
static int
seg_bio_add_bh(struct bio *bio, struct buffer_head *bh)
{
	unsigned int offset = ((unsigned long)bh->b_data & (PAGE_SIZE - 1));
	unsigned int len;
	request_queue_t *q = bdev_get_queue(bio->bi_bdev);

	BUG_ON(bio_flagged(bio, BIO_CLONED));
	/* Cloned bio must not modify io-vectors.
	   This is a check against future changes */

	if (SC_BYTES_SECTOR(bio->bi_size + bh->b_size) > q->max_sectors)
		return 1; /* FULL */

	if (bio->bi_vcnt > 0) {
		struct bio_vec *bvec = &bio->bi_io_vec[bio->bi_vcnt - 1];

		if (bvec->bv_page == bh->b_page &&
		    bvec->bv_offset + bvec->bv_len == offset) {
			bvec->bv_len += bh->b_size;
			goto out;
		}
	}
	len = bio_add_page(bio, bh->b_page, bh->b_size, offset);
	if (len == 0)
		return 1;
 out:
	get_bh(bh);
	return 0;
}

/*
 * Buffer array operations
 */
static inline int
ba_get_new_segsum_block(struct nilfs_sc_info *sci)
{
	struct buffer_head *bh;

	bh = nilfs_getblk(sci->sc_sb,
			  sci->sc_pseg_start + (sci->sc_ba_max - sci->sc_ba_sum_idx));
	if(!bh)
		return -ENOMEM;

	SetPageWriteback(bh->b_page);
	seg_debug(3, "adding bh(%p) to bh_arr[%d]\n",
		  bh, sci->sc_ba_sum_idx - 1);
	SC_BH(sci, --sci->sc_ba_sum_idx) = bh;
	return 0;
}

static inline void
ba_register_block(struct nilfs_sc_info *sci, struct buffer_head *bh)
{
	struct buffer_head **pbh = &SC_BH(sci, sci->sc_ba_idx);
	struct page *page = bh->b_page;

	BUG_ON(*pbh);

	lock_page(page);
	nilfs_set_page_writeback(page);
	seg_debug(3, "adding bh(%p) to bh_arr[%d]\n",
		  bh, pbh - &SC_BH(sci, 0));
	get_bh(bh);
	*pbh = bh;
	sci->sc_ba_idx++;
	sci->sc_sum.nblocks++;
	unlock_page(page);
}

static int ba_init(struct nilfs_sc_info *sci, struct nilfs_sb_info *sbi)
{
	int err;

	sci->sc_ba_idx = 0;
	sci->sc_ba_sum_idx = sci->sc_ba_max;

	err = ba_get_new_segsum_block(sci);
	if (err)
		return err;

	sci->sc_sum_bytes = sizeof(struct nilfs_seg_summary);
	sci->sc_sum.nblocks = sci->sc_sum.nblk_sum = 1;
	sci->sc_sum.nfinfo = sci->sc_sum.nfbinfo = 0;
	sci->sc_sum.nblk_file = sci->sc_sum.nblk_fbt = 0;
	sci->sc_sum.nblk_inode = sci->sc_sum.nblk_ibt = 0;
	sci->sc_sum.prev_pseg = sbi->s_prev_pseg;
	sci->sc_sum.seg_seq = sbi->s_nilfs->ns_seg_seq;
	sci->sc_blk_cnt = 0;

	return 0;
}

static void ba_clear(struct nilfs_sc_info *sci, int err)
{
	struct buffer_head **pbh = &SC_SUM_BH(sci, 0);
	struct buffer_head **ep = &SC_SUM_BH(sci, sci->sc_sum.nblk_sum);

	while (pbh > ep) {
		ClearPageWriteback((*pbh)->b_page);
		nilfs_brelse(*pbh);
		*pbh-- = NULL;
	}

	pbh = &SC_BH(sci, 0);
	ep = &SC_BH(sci, sci->sc_ba_idx);
	while (pbh < ep) {
		struct page *page = (*pbh)->b_page;

		if (PageWriteback(page)) {
			if (NILFS_SEG_ERR(err))
				SetPageError(page);
			else
				ClearPageError(page);
			nilfs_end_page_writeback(page);
		}
#ifdef CONFIG_NILFS_DEBUG
		else {
			/*
			 * Code for analysing unexpected BUG() at 
			 * end_page_writeback()
			 *
			 * NOTE: If the buffer size does not equal to
			 *   the page size, this happens as normal.
			 *   So, this code should be removed in the future.
			 */
			nilfs_warning(sci->sc_sb, __FUNCTION__,
				      "found page without writeback flag."
				      "(nblocks=%u, nblk_sum=%u, flag=0x%x, "
				      "ba_idx=%u, i=%u)\n",
				      sci->sc_sum.nblocks,
				      sci->sc_sum.nblk_sum,
				      sci->sc_sum.flags,
				      sci->sc_ba_idx,
				      pbh - sci->sc_bh_arr);
		}
#endif
		nilfs_brelse(*pbh);
		*pbh++ = NULL;
	}
}

/**
 * ba_required_segsum_block - get the number of additional summary blocks
 * @sci: nilfs_sc_info
 * @bytes: byte count of the segment information that will be appended.
 *
 * ba_required_segsum_block() returns how many additional blocks are 
 * required for appending specified bytes to the segment summary.
 * The segment header MUST be written in advance because an incorrect
 * result will be returned when sci->sc_sum_bytes == 0.
 */
static inline unsigned int
ba_required_segsum_block(struct nilfs_sc_info *sci, unsigned int bytes)
{
	return (((sci->sc_sum_bytes - 1) & (sci->sc_sb->s_blocksize - 1)) + bytes)
		>> sci->sc_sb->s_blocksize_bits;
}

/**
 * ba_append_to_segsum - append given data to the segment summary.
 * @sci: nilfs_sc_info
 * @vp: pointer to the data
 * @bytes: byte count of the data
 *
 * ba_append_to_segsum() appends specified data to the segment summary
 * holded by @sci. It extends the segment summary and copies segment 
 * information from @vp to the extended region. The segment summary
 * block are added as needed. The data can be copied over two blocks
 * but cannot exceed two blocks.
 * The copying of data can be skipped; if @vp is specified to NULL,
 * only the space required to store data is reserved and no data are 
 * written in.
 *
 * Return Value: On success, 0 is returned. on error, a following negative
 * error code is returned.
 * 
 * %-ENOMEM - Insufficient memory available.
 */
static int
ba_append_to_segsum(struct nilfs_sc_info *sci, void *vp, unsigned int bytes)
{
	unsigned long offset = sci->sc_sum_bytes & (sci->sc_sb->s_blocksize - 1);
	struct buffer_head *bh = SC_BH(sci, sci->sc_ba_sum_idx);
	unsigned int bcnt = 0;
	char *bp = vp;
	int err;

	BUG_ON(bytes > sci->sc_sb->s_blocksize);
	if (offset) {
		bcnt = min_t(unsigned int,
			     bytes, sci->sc_sb->s_blocksize - offset);
		if (bp)
			memcpy(bh->b_data + offset, bp, bcnt);

		if (bcnt == bytes)
			goto out;
	}
	/* Extending segment summary block */
	seg_debug(2, "extending summary block (bcnt=%u)\n", bcnt);
	err = ba_get_new_segsum_block(sci);
	if (err)
		return err;
	sci->sc_sum.nblocks++;
	sci->sc_sum.nblk_sum++;

	if (bp) {
		bp += bcnt;
		memcpy(bh->b_data, bp, bytes - bcnt);
	}
 out:
	sci->sc_sum_bytes += bytes;
	return 0;
}

static void
ba_write_to_segsum(struct nilfs_sc_info *sci, unsigned long offset,
		   void *p, unsigned int bytes)
{
	unsigned int b_index;
	unsigned int b_offset;
	struct buffer_head *bh;
	unsigned int bcnt;

	BUG_ON(offset + bytes > sci->sc_sum_bytes ||
	       bytes > sci->sc_sb->s_blocksize);
	b_index = sci->sc_ba_max - (offset >> sci->sc_sb->s_blocksize_bits) - 1;
	b_offset = offset & (sci->sc_sb->s_blocksize - 1);
	bh = SC_BH(sci, b_index);
	BUG_ON(bh == NULL);

	bcnt = min_t(unsigned int, bytes, sci->sc_sb->s_blocksize - b_offset);
	memcpy(bh->b_data + b_offset, p, bcnt);
	if (likely(bcnt == bytes))
		return;

	bh = SC_BH(sci, b_index - 1);
	BUG_ON(bh == NULL);

	memcpy(bh->b_data, (char *)p + bcnt, bytes - bcnt);
}

static void
ba_end_fdata(struct nilfs_sc_info *sci, struct inode *inode)
{
	struct nilfs_finfo finfo;
	struct nilfs_inode *raw_inode = nilfs_raw_inode(NILFS_I(inode));

	if (sci->sc_blk_cnt == 0)
		return;

	finfo.fi_nblocks = cpu_to_le32(sci->sc_blk_cnt);
	finfo.fi_version = raw_inode->i_version;  /* inode version for GC */
	finfo.fi_ino = cpu_to_le64(inode->i_ino);

	ba_write_to_segsum(sci, sci->sc_info_offset, &finfo, sizeof(finfo));

	sci->sc_sum.nblk_file += sci->sc_blk_cnt;
	sci->sc_sum.nfinfo++;
	sci->sc_blk_cnt = 0;
	sci->sc_dirty_file_ptr = NILFS_I(inode);
}

static inline int
ba_add_fdata(struct nilfs_sc_info *sci, struct inode *inode,
	     struct buffer_head *bh, fbn_t fbn)
{
	unsigned int required;
	unsigned long bytes = sizeof(__le64);
	int err;
	__le64 raw_fbn;
	
	if (sci->sc_blk_cnt == 0)
		bytes += sizeof(struct nilfs_finfo);

	required = ba_required_segsum_block(sci, bytes);
	if (sci->sc_sum.nblocks + required + 1 > sci->sc_residual_blocks) {
		ba_end_fdata(sci, inode);
		return NILFS_SEG_FULL;
	}

	/* Appending file block number of this block to segument summary */
	if (sci->sc_blk_cnt == 0) {
		sci->sc_info_offset = sci->sc_sum_bytes;
		err = ba_append_to_segsum(sci, NULL, sizeof(struct nilfs_finfo));
		if (err)
			goto failed;
	}
	raw_fbn = cpu_to_le64(fbn);
	err = ba_append_to_segsum(sci, &raw_fbn, sizeof(raw_fbn));
	if (err)
		goto failed;

	ba_register_block(sci, bh);
	sci->sc_blk_cnt++;
 failed:
	return err;
}

static void
ba_end_fbt_node(struct nilfs_sc_info *sci, struct inode *inode)
{
	struct nilfs_fbinfo fbinfo;

	if (sci->sc_blk_cnt == 0)
		return;

	fbinfo.fbi_nblocks = cpu_to_le32(sci->sc_blk_cnt);
	fbinfo.fbi_ino = cpu_to_le64(inode->i_ino);

	ba_write_to_segsum(sci, sci->sc_info_offset, &fbinfo, sizeof(fbinfo));

	sci->sc_sum.nblk_fbt += sci->sc_blk_cnt;
	sci->sc_sum.nfbinfo++;
	sci->sc_blk_cnt = 0;
}

static int
ba_add_fbt_node(struct nilfs_sc_info *sci, struct inode *inode,
		struct buffer_head *bh)
{
	unsigned int required;  /* additional blocks required */
	unsigned long bytes = sizeof(struct nilfs_fbinfo);
	int err = 0;
	
	if (sci->sc_blk_cnt == 0) {
		required = ba_required_segsum_block(sci, bytes);
		if (sci->sc_sum.nblocks + required + 1 > sci->sc_residual_blocks)
			return NILFS_SEG_FULL;

		sci->sc_info_offset = sci->sc_sum_bytes;
		err = ba_append_to_segsum(sci, NULL, sizeof(struct nilfs_fbinfo));
		if (err)
			goto failed;
	} else {
		if (sci->sc_sum.nblocks + 1 > sci->sc_residual_blocks) {
			ba_end_fbt_node(sci, inode);
			return NILFS_SEG_FULL;
		}
	}
	ba_register_block(sci, bh);
	sci->sc_blk_cnt++;
 failed:
	return err;
}

static inline int
ba_add_iblock(struct nilfs_sc_info *sci, struct buffer_head *bh, fbn_t fbn)
{
	struct nilfs_inode_hdr *ih = (struct nilfs_inode_hdr *)bh->b_data;
	struct nilfs_iinfo iinfo;
	unsigned int required;
	int err;
	
	required = ba_required_segsum_block(sci, sizeof(iinfo));
	if (sci->sc_sum.nblocks + required + 1 > sci->sc_residual_blocks) {
		return NILFS_SEG_FULL;
	}
	iinfo.ii_ino = ih->ih_ino;
	err = ba_append_to_segsum(sci, &iinfo, sizeof(iinfo));
	if (err)
		goto failed;

	ba_register_block(sci, bh);
	sci->sc_sum.nblk_inode++;
 failed:
	return err;
}

static inline int
ba_add_ibt_node(struct nilfs_sc_info *sci, struct buffer_head *bh)
{
	if (sci->sc_sum.nblocks + 1 > sci->sc_residual_blocks)
		return NILFS_SEG_FULL;

	ba_register_block(sci, bh);
	sci->sc_sum.nblk_ibt++;
	return 0;
}

static inline int
ba_add_cp(struct nilfs_sc_info *sci, unsigned int flags)
{
	dbn_t cp_start;
	struct buffer_head *bh;

	sci->sc_cp.flags = flags;

	if (sci->sc_sum.nblocks + 1 > sci->sc_residual_blocks)
		return NILFS_SEG_FULL;

	sci->sc_ba_cp_idx = sci->sc_ba_idx;
	/* This initialization must be done before calling
	   ba_write_to_cp() */

	/* Allocating checkpoint blocks */
	cp_start = sci->sc_pseg_start + sci->sc_ba_cp_idx + sci->sc_sum.nblk_sum;
	bh = nilfs_getblk(sci->sc_sb, cp_start);
	if (!bh)
		return -ENOMEM;
	SetPageWriteback(bh->b_page);
	SC_BH(sci, sci->sc_ba_idx) = bh;
	sci->sc_ba_idx++;
	sci->sc_sum.nblocks++;
	return 0;
}
/*
 * Callback functions that enumerate, mark, and collect dirty blocks
 */
static int 
mark_fbt_from_data(struct buffer_head *bh, struct inode *inode, fbn_t fbn)
{
	struct super_block *sb = inode->i_sb;
	struct nilfs_inode_info *ii = NILFS_I(inode);
	struct nilfs_sc_info *sci = NILFS_SC(NILFS_SB(sb));
	int err;

	if (!buffer_dirty(bh))
		return 0;

	err = nilfs_btree_mark_from_data(&ii->i_block_root, bh);
	if(err < 0) {
		/* error handling should be detailed */
		return err;
	}
	err = ba_add_fdata(sci, inode, bh, fbn);
	return err;
}

static int 
mark_fbt_from_node(struct buffer_head *bh, struct inode *inode)
{
	struct nilfs_inode_info *ii = NILFS_I(inode);
	int err;

	if (!buffer_dirty(bh) || buffer_prepare_dirty(bh))
		return 0;

	err = nilfs_btree_mark_from_node(&ii->i_block_root, bh);
	if(err < 0) {
		return err;
		/* error handling should be detailed */
	}
	return 0;
}

static int 
count_fbt_dirty(struct buffer_head *bh, struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct nilfs_sc_info *sci = NILFS_SC(NILFS_SB(sb));
	int err;

	BUG_ON(!buffer_dirty(bh));

	err = ba_add_fbt_node(sci, inode, bh);
	return err;
}

static int 
mark_ibt_from_data(struct buffer_head *bh, struct nilfs_sb_info *sbi, fbn_t fbn)
{
	struct nilfs_sc_info *sci = NILFS_SC(sbi);
	int err;
	
	if (!buffer_dirty(bh))
		return 0;

	seg_debug(2, "detected dirty inode block (fbn=%lu, 1st-ino=%llu)\n",
		  fbn,
		  le64_to_cpu(((struct nilfs_inode_hdr *)bh->b_data)->ih_ino));
	err = nilfs_btree_mark_from_data(&sbi->s_inode_root, bh);
	if(err < 0) {
		/* error handling should be detailed */
		return err;
	}
	err = ba_add_iblock(sci, bh, fbn);
	return err;
}

static int 
mark_ibt_from_node(struct buffer_head *bh, struct nilfs_sb_info *sbi)
{
	int err;

	if (!buffer_dirty(bh) || buffer_prepare_dirty(bh))
		return 0;

	err = nilfs_btree_mark_from_node(&sbi->s_inode_root, bh);
	if(err < 0) {
		return err;
		/* error handling should be detailed */
	}
	return 0;
}

static int 
count_ibt_dirty(struct buffer_head *bh, struct nilfs_sb_info *sbi)
{
	int err;

	BUG_ON(!buffer_dirty(bh));

	err = ba_add_ibt_node(NILFS_SC(sbi), bh);
	return err;
}

static int
lookup_dirty_buffers(struct super_block *sb, struct address_space *mapping,
		     void *private, data_proc_t proc)
{
	struct page *pages[SC_N_PAGEVEC];
	unsigned int nr_page;
	pgoff_t index = 0;
	int i;
	int err = 0;
	
 repeat:
#if NEED_RWLOCK_FOR_PAGECACHE_LOCK
	read_lock_irq(&mapping->tree_lock);
#else
	spin_lock_irq(&mapping->tree_lock);
#endif
	nr_page = radix_tree_gang_lookup_tag(&mapping->page_tree,
					     (void **)pages,
					     index,
					     SC_N_PAGEVEC,
					     PAGECACHE_TAG_DIRTY);
	for (i = 0; i < nr_page; i++)
		page_cache_get(pages[i]);
#if NEED_RWLOCK_FOR_PAGECACHE_LOCK
	read_unlock_irq(&mapping->tree_lock);
#else
	spin_unlock_irq(&mapping->tree_lock);
#endif

	if (nr_page == 0)
		return 0;
	index = pages[nr_page - 1]->index + 1;

	for(i = 0; i < nr_page; i++) {
		struct buffer_head *bh, *head;
		fbn_t fbn;

		if (err)
			goto skip_page;
		BUG_ON(!page_has_buffers(pages[i]));
		bh = head = page_buffers(pages[i]);
		fbn = pages[i]->index >> (PAGE_SHIFT - sb->s_blocksize_bits);
		do {
			if(buffer_dirty(bh)) {
				get_bh(bh);
				err = (*proc)(bh, private, fbn);
				put_bh(bh);
				if (err) break;
			}
			bh = bh->b_this_page;
			fbn++;
		} while (bh != head);
	skip_page:
		page_cache_release(pages[i]);
	}
	if (!err)
		goto repeat;

	return err;
}
	    
static int
lookup_buffers_64(struct radix_tree_64_root *tree, spinlock_t *tree_lock,
		  void *private, node_proc_t proc, int tag)
{
	struct page *pages[SC_N_PAGEVEC];
	unsigned int nr_page;
	u64 index = 0;
	int i;
	int err = 0;
	
 repeat:
	spin_lock_irq(tree_lock);
	nr_page = radix_tree_64_gang_lookup_tag(tree,
						(void **)pages,
						index,
						SC_N_PAGEVEC,
						tag);
	for (i = 0; i < nr_page; i++)
		page_cache_get(pages[i]);
	spin_unlock_irq(tree_lock);

	if (nr_page == 0)
		return 0;
	index = nilfs_node_page_index(pages[nr_page - 1]) + 1;

	for(i = 0; i < nr_page; i++) {
		struct buffer_head *bh, *head;

		if (err)
			goto skip_page;
		BUG_ON(!page_has_buffers(pages[i]));
		bh = head = page_buffers(pages[i]);
		do {
			get_bh(bh);
			err = (*proc)(bh, private);
			put_bh(bh);
			if (err) break;
			bh = bh->b_this_page;
		} while (bh != head);
	skip_page:
		page_cache_release(pages[i]);
	}
	if(!err)
		goto repeat;

	return err;
}

static inline int
btree_lookup_dirty_buffers(struct nilfs_btree *btree, void *private,
			   node_proc_t proc)
{
	struct buffer_head *bhs[SC_N_BHVEC];
	int nitems, i, err = 0;

	nilfs_btree_lookup_dirty_buffers_begin(btree);
 repeat:
	nitems = nilfs_btree_lookup_dirty_buffers(btree, bhs, SC_N_BHVEC);
	if (nitems == 0) {
		nilfs_btree_lookup_dirty_buffers_end(btree);
		return 0;
	}
	for (i = 0; i < nitems; i++) {
		if (likely(!err))
			err = (*proc)(bhs[i], private);
		nilfs_brelse(bhs[i]);
	}
	if (!err)
		goto repeat;

	nilfs_btree_lookup_dirty_buffers_end(btree);
	return err;
}

static void dispose_garbage_list(struct nilfs_sb_info *sbi)
{

 repeat:
	spin_lock_irq(&sbi->s_segctor.dirty_files_lock);
	if (!list_empty(&sbi->s_segctor.garbage_list)) {
		struct list_head *h = sbi->s_segctor.garbage_list.next;
		struct nilfs_inode_info *ii = list_entry(h, struct nilfs_inode_info, i_dirty);

		list_del_init(h);
		seg_debug(3, "deleting file (ino=%lu) from garbage_list\n",
			  ii->vfs_inode.i_ino);
		ii->i_state &= ~(NILFS_STATE_DIRTY | NILFS_STATE_COLLECTED | NILFS_STATE_UPDATED);
		brelse(ii->i_bh);
		ii->i_bh = NULL;

		spin_unlock_irq(&sbi->s_segctor.dirty_files_lock);
		iput(&ii->vfs_inode);
		goto repeat;
	}
	spin_unlock_irq(&sbi->s_segctor.dirty_files_lock);
}

/* 
 * reconfirm_construction - reconfirm needs of segment construction
 *        with clearing the global dirty flag of inode B-tree.
 *
 * This function must be called after synching with BIO completion routine.
 */
static inline int
reconfirm_construction(struct nilfs_sb_info *sbi)
{
	if (nilfs_btree_test_and_clear_dirty(&sbi->s_inode_root) ||
	    test_bit(NILFS_SC_IBLK_DIRTY, &sbi->s_segctor.flags)) {
		clear_bit(NILFS_SC_IBLK_DIRTY, &sbi->s_segctor.flags);
		set_bit(NILFS_SC_DIRTY, &sbi->s_segctor.flags);
	}

	if (list_empty(&sbi->s_segctor.dirty_files) &&
	    !test_bit(NILFS_SC_DIRTY, &sbi->s_segctor.flags)) {
		seg_debug(1, "Aborted construction (no changes found in "
			  "reconfirmation)\n");
		return NILFS_SEG_EMPTY;
	}
	return 0;
}

static inline void fill_in_file_block_roots(struct nilfs_sb_info *sbi)
{
	struct nilfs_inode_info *ii;
	struct nilfs_inode *raw_inode;

	/* spin_lock_irq(&sbi->s_segctor.dirty_files_lock); */
	list_for_each_entry(ii, &sbi->s_segctor.dirty_files, i_dirty) {
		raw_inode = nilfs_raw_inode(ii);
		raw_inode->i_block_root = cpu_to_le64(NILFS_BTREE_ROOT_DBN(&ii->i_block_root));
		ii->i_state = (ii->i_state & ~NILFS_STATE_DIRTY) | NILFS_STATE_COLLECTED;
	}
	/* spin_unlock_irq(&sbi->s_segctor.dirty_files_lock); */
}

/*
 * CRC calculation routines
 */
static inline void fill_in_segsum_crc(struct nilfs_sc_info *sci, u32 seed)
{
	unsigned long size, bytes = sci->sc_sum_bytes;
	struct buffer_head **pbh = &SC_SUM_BH(sci, 0);
	struct buffer_head **ep = &SC_SUM_BH(sci, sci->sc_sum.nblk_sum);
	struct nilfs_seg_summary *raw_sum;
	u32 crc;

	raw_sum = (struct nilfs_seg_summary *)((*pbh)->b_data);
	size = min_t(unsigned long, bytes, (*pbh)->b_size);
	crc = crc32c_le(seed,
			(unsigned char *)raw_sum +
			sizeof(raw_sum->ss_datasum) +
			sizeof(raw_sum->ss_sumsum),
			size - (sizeof(raw_sum->ss_datasum) + 
				sizeof(raw_sum->ss_sumsum)));
	while (--pbh > ep) {
		bytes -= size;
		size = min_t(unsigned long, bytes, (*pbh)->b_size);
		crc = crc32c_le(crc, (*pbh)->b_data, size);
	}
	raw_sum->ss_sumsum = cpu_to_le32(crc);
}

static inline void fill_in_cp_crc(struct nilfs_sc_info *sci, u32 seed)
{
	struct buffer_head **pbh = &SC_BH(sci, sci->sc_ba_cp_idx);
	struct nilfs_checkpoint *raw_cp;
	u32 crc;

	raw_cp = (struct nilfs_checkpoint *)((*pbh)->b_data);
	BUG_ON(NILFS_CP_BYTES > (*pbh)->b_size);
	crc = crc32c_le(seed, (unsigned char *)raw_cp + sizeof(raw_cp->cp_sum),
			NILFS_CP_BYTES - sizeof(raw_cp->cp_sum));

	raw_cp->cp_sum = cpu_to_le32(crc);
}

static inline void fill_in_data_crc(struct nilfs_sc_info *sci, u32 seed)
{
	struct buffer_head **pbh = &SC_SUM_BH(sci, 0);
	struct buffer_head **ep = &SC_SUM_BH(sci, sci->sc_sum.nblk_sum);
	struct nilfs_seg_summary *raw_sum;
	u32 crc;

	raw_sum = (struct nilfs_seg_summary *)((*pbh)->b_data);
	crc = crc32c_le(seed,
			(unsigned char *)raw_sum + sizeof(raw_sum->ss_datasum),
			(*pbh)->b_size - sizeof(raw_sum->ss_datasum));
	while (--pbh > ep)
		crc = crc32c_le(crc, (*pbh)->b_data, (*pbh)->b_size);

	pbh = &SC_BH(sci, 0);
	ep = &SC_BH(sci, sci->sc_sum.nblk_file);
	while (pbh < ep) {
		void *kaddr = kmap_atomic((*pbh)->b_page, KM_USER0);
		unsigned long offset = nilfs_page_block_offset(*pbh);

		crc = crc32c_le(crc, kaddr + offset, (*pbh)->b_size);
		kunmap_atomic(kaddr, KM_USER0);
		pbh++;
	}
	ep = &SC_BH(sci, sci->sc_ba_idx);
	while (pbh < ep) {
		crc = crc32c_le(crc, (*pbh)->b_data, (*pbh)->b_size);
		pbh++;
	}
	raw_sum->ss_datasum = cpu_to_le32(crc);
}

static inline void 
fill_in_cp(struct nilfs_sc_info *sci, struct nilfs_sb_info *sbi)
{
	struct nilfs_checkpoint *raw_cp;

	sci->sc_cp.inode_root = NILFS_BTREE_ROOT_DBN(&sbi->s_inode_root);

	/* Writing checkpoint blocks */
	raw_cp = (struct nilfs_checkpoint *)(SC_BH(sci, sci->sc_ba_cp_idx)->b_data);
	/* raw_cp->cp_sum = 0; */
	raw_cp->cp_flags = cpu_to_le16(sci->sc_cp.flags);
	raw_cp->cp_bytes = cpu_to_le16(NILFS_CP_BYTES);
	raw_cp->cp_inode_root =	cpu_to_le64(sci->sc_cp.inode_root);
	raw_cp->cp_nblk_lseg = cpu_to_le64(sci->sc_nblk_lseg);
	raw_cp->cp_inodes_count = cpu_to_le64(atomic_read(&sbi->s_inodes_count));
	raw_cp->cp_blocks_count = cpu_to_le64(atomic_read(&sbi->s_blocks_count));
}

static inline void fill_in_segsum(struct nilfs_sc_info *sci)
{
	struct nilfs_seg_summary *raw_sum;
	struct timespec ctime = CURRENT_TIME;

	raw_sum = (struct nilfs_seg_summary *)SC_SUM_BH(sci, 0)->b_data;
	raw_sum->ss_seq     = cpu_to_le64(sci->sc_sum.seg_seq);
	raw_sum->ss_magic   = cpu_to_le16(NILFS_SEGSUM_MAGIC);
	raw_sum->ss_nblocks = cpu_to_le16(sci->sc_sum.nblocks);
	raw_sum->ss_nfinfo  = cpu_to_le16(sci->sc_sum.nfinfo);
	raw_sum->ss_nfbinfo = cpu_to_le16(sci->sc_sum.nfbinfo);
	raw_sum->ss_nfblk   = cpu_to_le16(sci->sc_sum.nblk_file);
	raw_sum->ss_nfbblk  = cpu_to_le16(sci->sc_sum.nblk_fbt);
	raw_sum->ss_niblk   = cpu_to_le16(sci->sc_sum.nblk_inode);
	raw_sum->ss_nibblk  = cpu_to_le16(sci->sc_sum.nblk_ibt);
	raw_sum->ss_create  = cpu_to_le64(ctime.tv_sec);
	raw_sum->ss_prev    = cpu_to_le16(sci->sc_sum.prev_pseg);
	raw_sum->ss_flags   = cpu_to_le16(sci->sc_sum.flags);
}

static void redirty_files(struct nilfs_sb_info *sbi)
{
	struct nilfs_inode_info *ii;
	unsigned long flags;

	spin_lock_irqsave(&sbi->s_segctor.dirty_files_lock, flags);
	list_for_each_entry(ii, &sbi->s_segctor.dirty_files, i_dirty) {
		if (ii->i_state & NILFS_STATE_COLLECTED) {
			ii->i_state = (ii->i_state & ~NILFS_STATE_COLLECTED) |
				NILFS_STATE_DIRTY;
			seg_debug(2, "redirtied file (ino=%lu)\n",
				  ii->vfs_inode.i_ino);
		}
	}
	spin_unlock_irqrestore(&sbi->s_segctor.dirty_files_lock, flags);
}

static void drop_collected_files(struct nilfs_sb_info *sbi)
{
	struct list_head *head;
	struct nilfs_inode_info *ii, *n;
	unsigned long flags;

	head = &sbi->s_segctor.dirty_files;
	spin_lock_irqsave(&sbi->s_segctor.dirty_files_lock, flags);
	list_for_each_entry_safe(ii, n, head, i_dirty) {
		if (!(ii->i_state & NILFS_STATE_COLLECTED))
			continue;
		ii->i_state &= ~NILFS_STATE_COLLECTED;
		if (ii->i_state & NILFS_STATE_DIRTY)
			continue;

		list_del_init(&ii->i_dirty);
		seg_debug(3, "dropping collected file (ino=%lu)\n",
			  ii->vfs_inode.i_ino);
		ii->i_state |= NILFS_STATE_UPDATED;
		list_add_tail(&ii->i_dirty, &sbi->s_segctor.garbage_list);
	}
	spin_unlock_irqrestore(&sbi->s_segctor.dirty_files_lock, flags);
}

static inline int 
collect_segment_blocks(struct nilfs_sc_info *sci, struct nilfs_sb_info *sbi)
{
	struct list_head *head;
	struct nilfs_inode_info *ii;
	struct inode *inode;
	int err = 0;

	switch (sci->sc_stage) {
	case SC_COLLECT_INIT:
		seg_debug(2, "** LSEG BEGIN\n");
		/*
		 * Pre-processes before first segment construction are
		 * inserted here.
		 */
		sci->sc_sum.flags = SS_LOGBGN;
		sci->sc_dirty_file_ptr = NULL;
		sci->sc_stage++;
	case SC_COLLECT_DATA:
		seg_debug(2, "** DATA STAGE\n");
		/* spin_lock_irq(&sbi->s_segctor.dirty_files_lock); */
		head = &sbi->s_segctor.dirty_files;
		ii = list_prepare_entry(sci->sc_dirty_file_ptr, head, i_dirty);
		list_for_each_entry_continue(ii, head, i_dirty) {
			if (!(ii->i_state & NILFS_STATE_DIRTY))
				continue;
			inode = &ii->vfs_inode;

			err = lookup_dirty_buffers(sci->sc_sb, inode->i_mapping,
						   inode,
						   (data_proc_t)mark_fbt_from_data);
			if (err) {
				sci->sc_dirty_file_ptr = 
					list_entry(ii->i_dirty.prev,
						   struct nilfs_inode_info,
						   i_dirty);
				/* spin_unlock_irq(&sbi->s_segctor.dirty_files_lock); */
				goto break_or_fail;
			}
			ba_end_fdata(sci, inode);
		}
		sci->sc_dirty_file_ptr = NULL;
		/* spin_unlock_irq(&sbi->s_segctor.dirty_files_lock); */
		sci->sc_stage++;
	case SC_COLLECT_FBT:
		seg_debug(2, "** FBT STAGE\n");
		/* spin_lock_irq(&sbi->s_segctor.dirty_files_lock); */
		head = &sbi->s_segctor.dirty_files;
		ii = list_prepare_entry(sci->sc_dirty_file_ptr, head, i_dirty);
		list_for_each_entry_continue(ii, head, i_dirty) {
			if (!(ii->i_state & NILFS_STATE_DIRTY))
				continue;
			inode = &ii->vfs_inode;

			err = lookup_buffers_64(&ii->i_block_ntree,
						&ii->i_block_ntree_lock,
						inode,
						(node_proc_t)mark_fbt_from_node, 
						PAGECACHE64_TAG_DIRTY);
			if (err) {
				sci->sc_dirty_file_ptr = 
					list_entry(ii->i_dirty.prev,
						   struct nilfs_inode_info,
						   i_dirty);
				/* spin_unlock_irq(&sbi->s_segctor.dirty_files_lock); */
				goto break_or_fail;
			}
			
			err = btree_lookup_dirty_buffers(&ii->i_block_root,
							 inode,
							 (node_proc_t)count_fbt_dirty);
			if (err) {
				sci->sc_dirty_file_ptr = 
					list_entry(ii->i_dirty.prev,
						   struct nilfs_inode_info,
						   i_dirty);
				/* spin_unlock_irq(&sbi->s_segctor.dirty_files_lock); */
				goto break_or_fail;
			}
			ba_end_fbt_node(sci, inode);
		}
		sci->sc_dirty_file_ptr = NULL;
		/* spin_unlock_irq(&sbi->s_segctor.dirty_files_lock); */
		sci->sc_stage++;
	case SC_COLLECT_IBLK:
		seg_debug(2, "** INODE STAGE\n");
		err = lookup_dirty_buffers(sci->sc_sb, sbi->s_mapping, sbi,
					   (data_proc_t)mark_ibt_from_data);
		if (err) break;
		sci->sc_stage++;
	case SC_COLLECT_IBT:
		seg_debug(2, "** IBT STAGE\n");
		err = lookup_buffers_64(&sbi->s_inode_ntree,
					&sbi->s_inode_ntree_lock,
					sbi,
					(node_proc_t)mark_ibt_from_node, 
					PAGECACHE64_TAG_DIRTY);
		if (err) break;

		err = btree_lookup_dirty_buffers(&sbi->s_inode_root,
						 sbi,
						 (node_proc_t)count_ibt_dirty);
		if (err) break;
		sci->sc_stage++;
	case SC_COLLECT_CP:
		seg_debug(2, "** CP STAGE\n");
		/* Appending Checkpoint */
		err = ba_add_cp(sci, NILFS_CP_MAJOR);
		if (err) break;
		sci->sc_stage++;
	case SC_COLLECT_DONE:
		/*
		 * Post processes after final segment construction
		 * can be inserted here.
		 */
		sci->sc_sum.flags |= SS_FJCP | SS_LOGEND;
		seg_debug(2, "** LSEG END\n");
		return 0;
	default:
		BUG();
	}
 break_or_fail:
	/* ba_cancel_blocks(sbi, sci, ...); */
	if (NILFS_SEG_ERR(err))
		seg_debug(2, "** ERROR(%d)\n", err);
	else
		seg_debug(2, "** SEG FEED\n");
	return err;
}

static inline int
follow_up_check(struct nilfs_sc_info *sci, struct nilfs_sb_info *sbi)
{
	BUG_ON(!NILFS_SEG_HAS_CP(&sci->sc_sum));

        /* Check whether the segment is empty or not */
	if (sci->sc_sum.nblocks == sci->sc_sum.nblk_sum + 1 &&
	    NILFS_SEG_SIMPLEX(&sci->sc_sum)) {
		clear_bit(NILFS_SC_DIRTY, &sbi->s_segctor.flags);
		seg_debug(1, "Aborted construction (no blocks were collected)\n");
		return NILFS_SEG_EMPTY;
	}
	return 0;
}

static int
update_dbn(struct nilfs_sc_info *sci, struct nilfs_btree *inode_root)
{
	struct buffer_head **pbh, *bh;
	struct inode *inode;
	int err = 0;
	dbn_t dbn;
	unsigned int cnt;

	seg_debug(2, "called\n");
	pbh = &SC_BH(sci, 0);
	dbn = sci->sc_pseg_start + sci->sc_sum.nblk_sum;

	cnt = sci->sc_sum.nblk_file;
	while (cnt-- > 0) {
		BUG_ON(buffer_nilfs_node(*pbh));

		bh = *pbh++;
		inode = bh->b_page->mapping->host;
		err = nilfs_btree_update_data_dbn(&NILFS_I(inode)->i_block_root,
						  bh, dbn);
		if (err)
			goto out;
		bh->b_blocknr = (sector_t)dbn++;
	}

	cnt = sci->sc_sum.nblk_fbt;
	while (cnt-- > 0) {
		BUG_ON(!buffer_nilfs_bbt_node(*pbh));

		inode = (*pbh)->b_page->mapping->host;
		err = nilfs_btree_update_node_dbn(&NILFS_I(inode)->i_block_root,
						 *pbh, dbn);
		if (err)
			goto out;
		err = nilfs_move_dirty_file_node_blk(inode, pbh, dbn);
		if (err)
			goto out;
		pbh++;  dbn++;
	}
	cnt = sci->sc_sum.nblk_inode;
	while (cnt-- > 0) {
		BUG_ON(buffer_nilfs_node(*pbh));

		bh = *pbh++;
		err = nilfs_btree_update_data_dbn(inode_root, bh, dbn);
		if (err)
			goto out;
		bh->b_blocknr = (sector_t)dbn++;
	}

	cnt = sci->sc_sum.nblk_ibt;
	while (cnt-- > 0) {
		BUG_ON(!buffer_nilfs_ibt_node(*pbh));

		err = nilfs_btree_update_node_dbn(inode_root, *pbh, dbn);
		if (err)
			goto out;
		err = nilfs_move_dirty_inode_node_blk(sci->sc_sb, pbh, dbn);
		if (err)
			goto out;
		pbh++;  dbn++;
	}
 out:
	seg_debug(2, "done\n");
	return err;
}

static void abort_segment_io(struct nilfs_sc_info *sci, int nblk_submit)
{
	struct buffer_head **pbh;
	struct page *page, *prev_page = NULL;
	int idx = sci->sc_sum.nblocks;

	seg_debug(2, "aborting segment (nblk_submit=%d)\n", nblk_submit);
	pbh = &SC_BH(sci, sci->sc_ba_idx);
	while (idx > sci->sc_sum.nblk_sum) {
		--idx;  --pbh;
		page = (*pbh)->b_page;
		if(idx < nblk_submit) {
			/* clear_buffer_uptodate(*pbh); */
			unlock_buffer(*pbh);
			if (page != prev_page) {
				if (prev_page)
					unlock_page(prev_page);
				prev_page = page;
			}
		}
		if (PageWriteback(page)) {
			SetPageError(page);
			nilfs_end_page_writeback(page);
		}
		nilfs_brelse_isr(sci->sc_sb, *pbh);
		*pbh = NULL;
	}
	if (prev_page)
		unlock_page(prev_page);

	prev_page = NULL;
	pbh = &SC_SUM_BH(sci, sci->sc_sum.nblk_sum);
	while (idx > 0) {
		--idx;  ++pbh;
		page = (*pbh)->b_page;
		if(idx < nblk_submit) {
			/* clear_buffer_uptodate(*pbh); */
			unlock_buffer(*pbh);
			if (page != prev_page) {
				if (prev_page)
					unlock_page(prev_page);
				prev_page = page;
			}
		}
		ClearPageWriteback(page);
		nilfs_brelse_isr(sci->sc_sb, *pbh);
		*pbh = NULL;
	}
	if (prev_page)
		unlock_page(prev_page);

	if (sci->sc_prev_stage < SC_COLLECT_IBLK &&
	    sci->sc_stage >= SC_COLLECT_IBLK)
		redirty_files(NILFS_SB(sci->sc_sb));
}

static void complete_segment_io(struct nilfs_sc_info *sci)
{
	struct nilfs_sb_info *sbi = NILFS_SB(sci->sc_sb);
	struct buffer_head **pbh;
	struct page *page, *prev_page = NULL;
	int idx = sci->sc_sum.nblocks;

	/*
	 * We assume that the buffers which belong to the same page
	 * continue over the buffer head array.
	 * Under this assumption, the last BHs of pages is identifiable
	 * by the discontinuity of bh->b_page (page != prev_page).
	 */
	seg_debug(2, "completing segment\n");
	pbh = &SC_BH(sci, sci->sc_ba_idx);
	while (idx > sci->sc_sum.nblk_sum) {
		--idx;  --pbh;
		set_buffer_uptodate(*pbh);
		clear_buffer_dirty(*pbh);
		unlock_buffer(*pbh);
		page = (*pbh)->b_page;
		if (page != prev_page) {
			if (prev_page) {
				if (!nilfs_page_buffers_dirty(prev_page))
					nilfs_clear_page_dirty(prev_page);
				unlock_page(prev_page);
			}
			prev_page = page;
		}
		if (PageWriteback(page)) {
			ClearPageError(page);
			nilfs_end_page_writeback(page);
		}
		nilfs_brelse_isr(sci->sc_sb, *pbh);
		*pbh = NULL;
	}
	if (prev_page) {
		if (!nilfs_page_buffers_dirty(prev_page))
			nilfs_clear_page_dirty(prev_page);
		unlock_page(prev_page);
	}	

	prev_page = NULL;
	pbh = &SC_SUM_BH(sci, sci->sc_sum.nblk_sum);
	while (idx > 0) {
		--idx;  ++pbh;
		set_buffer_uptodate(*pbh);
		clear_buffer_dirty(*pbh);
		unlock_buffer(*pbh);
		page = (*pbh)->b_page;
		if (page != prev_page) {
			if (prev_page) {
				if (!nilfs_page_buffers_dirty(prev_page))
					nilfs_clear_page_dirty(prev_page);
				unlock_page(prev_page);
			}
			prev_page = page;
		}
		ClearPageWriteback(page);
		nilfs_brelse_isr(sci->sc_sb, *pbh);
		*pbh = NULL;
	}
	if (prev_page) {
		if (!nilfs_page_buffers_dirty(prev_page))
			nilfs_clear_page_dirty(prev_page);
		unlock_page(prev_page);
	}
	drop_collected_files(sbi);
	if (NILFS_SEG_LOGEND(&sci->sc_sum)) {
		unsigned long flags;
		struct the_nilfs *nilfs = sbi->s_nilfs;

		/* Completed one logical segment */
		clear_bit(NILFS_SC_DIRTY, &sbi->s_segctor.flags);
		spin_lock_irqsave(&nilfs->ns_lock, flags);
		nilfs->ns_last_pseg_with_cp = sci->sc_pseg_start;
		nilfs->ns_inode_root = sci->sc_cp.inode_root;
		spin_unlock_irqrestore(&nilfs->ns_lock, flags);
		seg_debug(2, "completed logical segment\n");
	}

	/* Completed all BIOs of this segment */
	sbi->s_nilfs->ns_last_pseg = sci->sc_pseg_start;
	set_bit(NILFS_SC_LAST_CP, &sbi->s_segctor.flags);
}

static inline int 
__seg_bh_add_and_submit(struct nilfs_sc_info *sci,  struct nilfs_sb_info *sbi,
			struct bio **pbio, int idx, struct buffer_head *bh,
			int mode, int nr_vecs)
{
	int res;

	if (!*pbio) {
		*pbio = alloc_seg_bio(sci->sc_sb, sci->sc_pseg_start + idx,
				      nr_vecs);
		if (!*pbio)
			return -ENOMEM;
	}
	res = seg_bio_add_bh(*pbio, bh);
	if (res < 0) {
		clear_seg_bio(sbi, *pbio);
		bio_put(*pbio);
	} else if (res > 0) { /* bio is FULL (*pbh was not added to bio) */
		int err = submit_seg_bio(*pbio, sbi, mode);
		if (err)
			return err;
	}
	return res;
}

static int write_segment(struct nilfs_sc_info *sci, struct nilfs_sb_info *sbi)
{
	struct buffer_head **pbh, **ep;
	struct bio *bio = NULL;
	int si, idx, rest = sci->sc_sum.nblocks;
	int max_pages = bio_get_nr_vecs(sbi->s_super->s_bdev);
	int res;
	int rw = nilfs_test_opt(sbi, SYNC_BIO) ? WRITE_SYNC : WRITE;
	int has_cp = NILFS_SEG_HAS_CP(&sci->sc_sum);

	BUG_ON(sci->sc_nbio > 0);
	atomic_set(&sci->sc_bio_blk_cnt, rest);

	seg_debug(2, "submitting summary blocks (nblk_sum=%d)\n",
		  sci->sc_sum.nblk_sum);
	pbh = &SC_SUM_BH(sci, 0);
	ep = &SC_SUM_BH(sci, sci->sc_sum.nblk_sum);
	si = idx = 0;
	while(pbh > ep) {
		res = __seg_bh_add_and_submit(sci, sbi, &bio, idx,
					      *pbh, rw, min(max_pages, rest));
		if (res < 0)
			goto failed_bio;
		else if (res > 0) { /* bio is FULL (*pbh was not added to bio) */
			sci->sc_nbio++;
			rest -= idx - si;
			si = idx;
			bio = NULL;
		} else {
			lock_buffer(*pbh);
			idx++;  pbh--;
		}
	}

	seg_debug(2, "submitting normal blocks (ba_idx=%d, idx=%d)\n",
		  sci->sc_ba_idx, idx);
	pbh = &SC_BH(sci, 0);
	ep = &SC_BH(sci, sci->sc_ba_idx);
	if (has_cp)
		ep--;
	si = idx;
	while (pbh < ep) {
		res = __seg_bh_add_and_submit(sci, sbi, &bio, idx,
					      *pbh, rw, min(max_pages, rest));
		if (res < 0)
			goto failed_bio;
		else if (res > 0) { /* bio is FULL (*pbh was not added to bio) */
			sci->sc_nbio++;
			rest -= idx - si;
			si = idx;
			bio = NULL;
		} else {
			lock_buffer(*pbh);
			idx++;  pbh++;
		}
	}
	if (bio) {
		/*
		 * Last BIO is always sent through the following
		 * submission.
		 */
		if (!has_cp && nilfs_test_opt(sbi, SYNC_SEGMENT))
			rw |= (1 << BIO_RW_SYNC);
		res = submit_seg_bio(bio, sbi, rw);
		if (res)
			goto failed_bio;
		sci->sc_nbio++;
	}

	if (has_cp) {
		int barrier_done = 0;

		seg_debug(2, "submitting a checkpoint block (ba_idx=%d, idx=%d)\n",
			  sci->sc_ba_idx, idx);

		if (nilfs_test_opt(sbi, SYNC_SEGMENT))
			rw |= (1 << BIO_RW_SYNC);
		if (nilfs_test_opt(sbi, BARRIER)) {
			rw |= (1 << BIO_RW_BARRIER);
			barrier_done = 1;
		}
	retry:
		bio = NULL;
		res = __seg_bh_add_and_submit(sci, sbi, &bio, idx, *pbh, 0, 1);
		if (res < 0)
			goto failed_bio;
		else if (res > 0)
			BUG();

		lock_buffer(*pbh);
		idx++;  /* pbh++; */

		res = submit_seg_bio(bio, sbi, rw);
		if (res == -EOPNOTSUPP && barrier_done) {
			nilfs_warning(sbi->s_super, __FUNCTION__,
				      "barrier-based sync failed. "
				      "disabling barriers\n");
			nilfs_clear_opt(sbi, BARRIER);
			idx--;  /* pbh--; */
			unlock_buffer(*pbh);
			barrier_done = 0;
			rw &= ~(1 << BIO_RW_BARRIER);
			goto retry;
		} else if (res)
			goto failed_bio;
		sci->sc_nbio++;
	}

	seg_debug(1, "submitted a segment\n");
	return 0;

 failed_bio:
	seg_debug(1, "Failed to write. getting back the state of segment\n");
	set_bit(NILFS_SC_EIO, &sbi->s_segctor.flags);
	wait_on_segment(sbi, sci);
	abort_segment_io(sci, idx);
	return res;
}

static int 
__construct_segment(struct nilfs_sb_info *sbi, struct nilfs_sc_info *sci)
{
	struct the_nilfs *nilfs = sbi->s_nilfs;
	dbn_t seg_start, seg_end;
	int err;

	sci->sc_prev_stage = sci->sc_stage = 0;
	sci->sc_nblk_lseg = 0;

	do {
		nilfs_get_segment_range(sbi, nilfs->ns_last_segnum,
					&seg_start, &seg_end);

		/* Synchronizing previous BIOs */
		err = wait_on_segment(sbi, sci);
		if (err) {
			nilfs_invalidate_segment(sbi, seg_start, seg_end);
			goto out;
		}
		
		/* Re-check needs of construction */
		if (sci->sc_stage == SC_COLLECT_INIT) {
			err = reconfirm_construction(sbi);
			if (err)
				goto out;
		}

		/* Determining next (partial) segment */
		sci->sc_pseg_start = seg_start + sbi->s_pseg_offset;
		sci->sc_residual_blocks = (seg_end + 1) - sci->sc_pseg_start;
		if(sci->sc_residual_blocks < NILFS_PSEG_MIN_BLOCKS) {
			int segnum;
			
			/* Allocating a new full segment when required */
			err = nilfs_alloc_segment(sbi, &segnum, 1);
			if (err)
				goto out;

			nilfs_get_segment_range(sbi, segnum, &seg_start, &seg_end);
			nilfs->ns_last_segnum = segnum;
			nilfs->ns_seg_seq++;
			sbi->s_pseg_offset = 0;
			sbi->s_prev_pseg = 0;
			sci->sc_pseg_start = seg_start;
			sci->sc_residual_blocks = seg_end - seg_start + 1;
		}

		sci->sc_prev_stage = sci->sc_stage;
		err = ba_init(sci, sbi);
		if (err)
			goto out;

		err = collect_segment_blocks(sci, sbi);
		if (NILFS_SEG_ERR(err))
			goto failed_ba;

		if (sci->sc_stage == SC_COLLECT_DONE) {
			err = follow_up_check(sci, sbi);
			if (err)
				goto failed_ba;
		}

		/* Update hint about the current logical segment */
		sci->sc_nblk_lseg += sci->sc_sum.nblocks;

#ifdef CONFIG_NILFS_DEBUG
		if (nilfs_debug_info.verbose[NILFS_VERBOSE_SEGINFO])
			print_segment_info(sci, seg_start, seg_end);
#endif

		err = update_dbn(sci, &sbi->s_inode_root);
		if (err)
			goto failed_ba;

		if (sci->sc_prev_stage < SC_COLLECT_IBLK &&
		    sci->sc_stage >= SC_COLLECT_IBLK)
			fill_in_file_block_roots(sbi);

		if (NILFS_SEG_HAS_CP(&sci->sc_sum))
			fill_in_cp(sci, sbi);

		fill_in_segsum(sci);

		/*
		 * Checksum calculations
		 */
		if (NILFS_SEG_HAS_CP(&sci->sc_sum))
			fill_in_cp_crc(sci, sbi->s_crc_seed);
		fill_in_segsum_crc(sci, sbi->s_crc_seed);
		fill_in_data_crc(sci, sbi->s_crc_seed);

		/* Writing back the segment */
		err = write_segment(sci, sbi);
		if (err)
			goto out; /* The buffer array has been cleared */

		/* Forwarding pointers */
		sbi->s_pseg_offset += sci->sc_sum.nblocks;
		sbi->s_prev_pseg = sci->sc_sum.nblocks;
		sci->sc_sum.flags &= ~SS_LOGBGN;

	} while (sci->sc_stage != SC_COLLECT_DONE);

	/* Wait for completion of segment write */
	err = wait_on_segment(sbi, sci);
	if (err)
		nilfs_invalidate_segment(sbi, seg_start, seg_end);

	seg_debug(1, "submitted all segments\n");
 out:
	return err;

 failed_ba:
	ba_clear(sci, err);
	goto out;
}

/**
 * nilfs_segctor_commit - set timer of background write
 * @sci: nilfs_sc_info
 *
 * nilfs_segctor_commit() set a timer which triggers a segment construction.
 * If the timer has already been set, it ignores the new request.
 * This function MUST be called within a section locking the segment semaphore.
 */
static void nilfs_segctor_commit(struct nilfs_sc_info *sci)
{
	BUG_ON(!sci);
	spin_lock(&sci->sc_state_lock);
	if (!(sci->sc_flags & NILFS_SEGCTOR_COMMIT)) {
		sci->sc_timer->expires = jiffies + sci->sc_interval;
		add_timer(sci->sc_timer);
		sci->sc_flags |= NILFS_SEGCTOR_COMMIT;
	}
	spin_unlock(&sci->sc_state_lock);
}

static void __nilfs_segctor_kick(struct nilfs_sc_info *sci)
{
	if (sci->sc_flags & NILFS_SEGCTOR_INIT)
		return;

	if (!(sci->sc_flags & NILFS_SEGCTOR_KICK)) {
		sci->sc_flags |= NILFS_SEGCTOR_KICK;
		wake_up(&sci->sc_wait_request);
	}
}

/**
 * nilfs_segctor_kick - trigger a segment construction.
 * @sci: nilfs_sc_info
 *
 * nilfs_segctor_kick() wakes up segctord and marks a flag to build a segment.
 */
void nilfs_segctor_kick(struct nilfs_sc_info *sci)
{
	BUG_ON(!sci);
	spin_lock(&sci->sc_state_lock);
	__nilfs_segctor_kick(sci);
	spin_unlock(&sci->sc_state_lock);
}

void nilfs_segctor_add_dirty(struct nilfs_sc_info *sci, unsigned delta)
{
	BUG_ON(!sci);
	spin_lock(&sci->sc_state_lock);
	sci->sc_nr_dirty += delta;
	if (sci->sc_nr_dirty > sci->sc_block_max)
		__nilfs_segctor_kick(sci);
	spin_unlock(&sci->sc_state_lock);
}

/**
 * nilfs_construct_segment - construct a logical segment
 * @sb: super block
 * @wbc: writeback_control structure
 *
 * nilfs_construct_segment() construct a segment. This thread base
 * implementation triggers a call of __construct_segment() by
 * the segment constructor daemon (segctord) and waits for its completion.
 * Non-blocking mode request or background write is always ignored.
 * 
 * Return Value: On success, one of following non-negative values is 
 * returned.
 *
 * 0 - a logical segment was constructed and was written out.
 *
 * %NILFS_SEG_EMPTY - the construction was skipped because no dirty block
 * was found.
 *
 * %NILFS_SEG_IGNORED - a background write request was ignored.
 *
 * On errors, one of following negative error code is returned.
 *
 * %-EROFS - Read only filesystem.
 *
 * %-EFAULT - Internal error (FS-shutdown is going?)
 *
 * %-EIO - I/O error
 *
 * %-ENOSPC - Disk full.
 *
 * %-ENOMEM - Insufficient memory available.
 */
int nilfs_construct_segment(struct super_block *sb,
			    struct writeback_control *wbc)
{
	struct nilfs_sb_info *sbi = NILFS_SB(sb);
	struct nilfs_sc_info *sci = NILFS_SC(sbi);
	struct nilfs_transaction_info *ti;
	int err = 0;
	__u32 seq;

	if ((sb->s_flags & MS_RDONLY) || !sci) {
		seg_debug(1, "Skipped construction (read only)\n");
		return -EROFS;
	}

	if (wbc->nonblocking || wbc->sync_mode == WB_SYNC_NONE)
		return NILFS_SEG_IGNORED; /* skip background write */

	/* A call inside transactions causes a deadlock. */
	BUG_ON((ti = current->journal_info) && ti->ti_super == sb);

	spin_lock(&sci->sc_state_lock);

	if (sci->sc_flags & NILFS_SEGCTOR_INIT) {
		err = -EROFS;
		goto out_unlock;
	}
	seq = ++sci->sc_seq_request;
	while (nilfs_cnt32_lt(sci->sc_seq_done, seq)) {
		DEFINE_WAIT(wait);
		int should_sleep = 1;

		wake_up(&sci->sc_wait_request);
		prepare_to_wait(&sci->sc_wait_done, &wait,
				TASK_INTERRUPTIBLE);

		if (nilfs_cnt32_ge(sci->sc_seq_done, seq))
			should_sleep = 0;

		if (should_sleep) {
			spin_unlock(&sci->sc_state_lock);
			schedule();
			spin_lock(&sci->sc_state_lock);
		}
		finish_wait(&sci->sc_wait_done, &wait);
	}
	err = sci->sc_errno;
	sci->sc_seq_ack++;
	wake_up(&sci->sc_wait_request);

 out_unlock:
	spin_unlock(&sci->sc_state_lock);
	return err;
}

static void nilfs_segctor_construct(struct nilfs_sc_info *sci)
{
	struct nilfs_sb_info *sbi = NILFS_SB(sci->sc_sb);
	__u32 seq_accepted;
	int err;

	/*
	 * Writing lock the segment semaphore.
	 * A writer can monopolies s_segctor.dirty_files without spinlock 
	 */
	might_sleep();
	down_write(&sbi->s_segctor.sem);

	BUG_ON(!sci || !sci->sc_bh_arr);
	if (sci->sc_timer)
		del_timer_sync(sci->sc_timer);
	spin_lock(&sci->sc_state_lock);
	seq_accepted = sci->sc_seq_request;
	spin_unlock(&sci->sc_state_lock);

	if (list_empty(&sbi->s_segctor.dirty_files) &&
	    !test_bit(NILFS_SC_DIRTY, &sbi->s_segctor.flags)) {
		/*
		 * Segment is constructed synchronously.
		 * So, all bio is completed here.
		 *
		 * We can safely move the dirty flag from B-tree.
		 */
		if (!nilfs_btree_test_and_clear_dirty(&sbi->s_inode_root) &&
		    !test_bit(NILFS_SC_IBLK_DIRTY, &sbi->s_segctor.flags)) {
			err = NILFS_SEG_EMPTY;
			seg_debug(1, "Skipped construction (no changes)\n");
			goto escape;
		}
		clear_bit(NILFS_SC_IBLK_DIRTY, &sbi->s_segctor.flags);
		set_bit(NILFS_SC_DIRTY, &sbi->s_segctor.flags);
	}

	seg_debug(1, "begin\n");
	err = __construct_segment(sbi, sci);
	seg_debug(1, "end (stage=%d)\n", sci->sc_stage);

 escape:
	spin_lock(&sci->sc_state_lock);
	sci->sc_seq_done = seq_accepted;
	sci->sc_flags &= ~(NILFS_SEGCTOR_COMMIT | NILFS_SEGCTOR_KICK);
	sci->sc_errno = err;
	sci->sc_nr_dirty = 0;
	spin_unlock(&sci->sc_state_lock);

	up_write(&sbi->s_segctor.sem);
	if (!NILFS_SEG_ERR(err))
		dispose_garbage_list(sbi);
}

static void construction_timeout(unsigned long data)
{
	struct task_struct *p = (struct task_struct *)data;
	wake_up_process(p);
}

/**
 * segctor_thread - main loop of the segment constructor thread.
 * @arg: pointer to a struct nilfs_sc_info.
 *
 * segctor_thread() initializes a timer and serves as a daemon
 * to execute segment constructions.
 */
static int segctor_thread(void *arg)
{
	struct nilfs_sc_info *sci = (struct nilfs_sc_info *)arg;
	struct timer_list timer;

	daemonize("segctord");

	init_timer(&timer);
	timer.data = (unsigned long)current;
	timer.function = construction_timeout;
	sci->sc_timer = &timer;

	/* start sync. */
	sci->sc_task = current;
	wake_up(&sci->sc_wait_done);
	printk(KERN_INFO
	       "segctord starting. Construction interval = %ld seconds\n",
	       sci->sc_interval / HZ);

	spin_lock(&sci->sc_state_lock);
 loop:
	if (sci->sc_flags & NILFS_SEGCTOR_QUIT)
		goto end_thread;

	seg_debug(2, "sequence: req=%u, done=%u, ack=%u\n",
		  sci->sc_seq_request, sci->sc_seq_done,
		  sci->sc_seq_ack);

	if (sci->sc_seq_ack == sci->sc_seq_done) {
		if((sci->sc_flags & (NILFS_SEGCTOR_COMMIT | NILFS_SEGCTOR_KICK)) ||
		   sci->sc_seq_request != sci->sc_seq_done) {
			spin_unlock(&sci->sc_state_lock);
			nilfs_segctor_construct(sci);
			spin_lock(&sci->sc_state_lock);
			goto loop;
		}
	}

	wake_up(&sci->sc_wait_done);
#if NEED_REFRIGERATOR_ARGS
	if (current->flags & PF_FREEZE) {
#else
	if (freezing(current)) {
#endif
		seg_debug(2, "suspending segctord\n");
		spin_unlock(&sci->sc_state_lock);
#if NEED_REFRIGERATOR_ARGS
		refrigerator(PF_FREEZE);
#else
		refrigerator();
#endif
		spin_lock(&sci->sc_state_lock);
	} else {
		DEFINE_WAIT(wait);
		int should_sleep = 1;

		prepare_to_wait(&sci->sc_wait_request, &wait,
				TASK_INTERRUPTIBLE);

		if (sci->sc_seq_ack == sci->sc_seq_done) {
			if (sci->sc_seq_request != sci->sc_seq_done)
				should_sleep = 0;
			else if (sci->sc_flags & NILFS_SEGCTOR_KICK)
				should_sleep = 0;
			else if (sci->sc_flags & NILFS_SEGCTOR_COMMIT)
				should_sleep =
					time_before(jiffies, 
						    sci->sc_timer->expires);
		}
		if (should_sleep) {
			spin_unlock(&sci->sc_state_lock);
			schedule();
			spin_lock(&sci->sc_state_lock);
		}
		finish_wait(&sci->sc_wait_request, &wait);
	}
	seg_debug(2, "woke %s\n", 
		  ((sci->sc_flags & NILFS_SEGCTOR_COMMIT) &&
		   time_after_eq(jiffies, sci->sc_timer->expires)) ?
		  "(timeout)" : "");
	goto loop;

 end_thread:
	spin_unlock(&sci->sc_state_lock);
	del_timer_sync(sci->sc_timer);
	sci->sc_timer = NULL;

	/* end sync. */
	sci->sc_task = NULL;
	wake_up(&sci->sc_wait_done);
	seg_debug(1, "segctord exiting.\n");
	return 0;
}

static void nilfs_start_segctor_thread(struct nilfs_sc_info *sci)
{
	kernel_thread(segctor_thread, sci,
		      CLONE_VM | CLONE_FS | CLONE_FILES);
	wait_event(sci->sc_wait_done, sci->sc_task != NULL);
}

static void nilfs_kill_segctor_thread(struct nilfs_sc_info *sci)
{
	sci->sc_flags |= NILFS_SEGCTOR_QUIT;

	while (sci->sc_task) {
		wake_up(&sci->sc_wait_request);
		spin_unlock(&sci->sc_state_lock);
		wait_event(sci->sc_wait_done, sci->sc_task == NULL);
		spin_lock(&sci->sc_state_lock);
	}
}

/**
 * nilfs_segctor_reset - reset the segment constructor thread.
 * @sci: nilfs_sc_info
 *
 * nilfs_segctor_reset() resets sequence counters used for handshaking
 * with a segment constructor thread (segctord), and starts segctord.
 */
int nilfs_segctor_reset(struct nilfs_sc_info *sci)
{
	sci->sc_flags &= ~NILFS_SEGCTOR_INIT;
	sci->sc_seq_done = sci->sc_seq_request;
	sci->sc_seq_ack = sci->sc_seq_done;
	sci->sc_nr_dirty = 0;
	nilfs_start_segctor_thread(sci);
	return 0;
}

static ssize_t
nilfs_segctor_show_interval(struct the_nilfs *nilfs, char *page)
{
	struct nilfs_sb_info *sbi = nilfs_get_writer(nilfs);
	struct nilfs_sc_info *sci;
	unsigned long val;
	ssize_t ret;

	if (!sbi || !(sci = NILFS_SC(sbi)))
		return 0;

	spin_lock(&sci->sc_state_lock);
	val = sci->sc_interval / HZ;
	spin_unlock(&sci->sc_state_lock);
	ret = sprintf(page, "%lu\n", val);
	return ret;
}

static ssize_t
nilfs_segctor_store_interval(struct the_nilfs *nilfs, const char *page,
			  size_t length)
{
	struct nilfs_sb_info *sbi = nilfs_get_writer(nilfs);
	struct nilfs_sc_info *sci;
	unsigned long val;
	ssize_t ret;

	if (!sbi || !(sci = NILFS_SC(sbi)))
		return -EINVAL;

	ret = sscanf(page, "%lu", &val);
	if (ret != 1)
		return -EINVAL;
	if (!val)
		val = NILFS_SC_DEFAULT_MAX_AGE;
	spin_lock(&sci->sc_state_lock);
	sci->sc_interval = val * HZ;
	spin_unlock(&sci->sc_state_lock);
	return length;
}

static ssize_t
nilfs_segctor_show_threshold(struct the_nilfs *nilfs, char *page)
{
	struct nilfs_sb_info *sbi = nilfs_get_writer(nilfs);
	struct nilfs_sc_info *sci;
	unsigned long val;
	ssize_t ret;

	if (!sbi || !(sci = NILFS_SC(sbi)))
		return 0;

	spin_lock(&sci->sc_state_lock);
	val = sci->sc_block_max;
	spin_unlock(&sci->sc_state_lock);
	ret = sprintf(page, "%lu\n", val);
	return ret;
}

static ssize_t
nilfs_segctor_store_threshold(struct the_nilfs *nilfs, const char *page,
			   size_t length)
{
	struct nilfs_sb_info *sbi = nilfs_get_writer(nilfs);
	struct nilfs_sc_info *sci;
	unsigned long val;
	ssize_t ret;

	if (!sbi || !(sci = NILFS_SC(sbi)))
		return -EINVAL;

	ret = sscanf(page, "%lu", &val);
	if (ret != 1)
		return -EINVAL;
	if (!val)
		val = NILFS_SC_DEFAULT_THRESHOLD_DATA_AMOUNT;

	spin_lock(&sci->sc_state_lock);
	sci->sc_block_max = val;
	if (sci->sc_nr_dirty > sci->sc_block_max)
		__nilfs_segctor_kick(sci);
	spin_unlock(&sci->sc_state_lock);
	return length;
}

static ssize_t
nilfs_segctor_show_max_bio(struct the_nilfs *nilfs, char *page)
{
	struct nilfs_sb_info *sbi = nilfs_get_writer(nilfs);
	struct nilfs_sc_info *sci;
	ssize_t ret;

	if (!sbi || !(sci = NILFS_SC(sbi)))
		return 0;
	ret = sprintf(page, "%u\n", atomic_read(&sci->sc_max_bio));
	return ret;
}

static ssize_t
nilfs_segctor_store_max_bio(struct the_nilfs *nilfs, const char *page,
			   size_t length)
{
	struct nilfs_sb_info *sbi = nilfs_get_writer(nilfs);
	struct nilfs_sc_info *sci;
	unsigned int val;
	ssize_t ret;

	if (!sbi || !(sci = NILFS_SC(sbi)))
		return -EINVAL;

	ret = sscanf(page, "%u", &val);
	if (ret != 1)
		return -EINVAL;
	if (!val)
		val = NILFS_SC_DEFAULT_MAX_BIO;

	atomic_set(&sci->sc_max_bio, val);
	return length;
}

struct nilfs_attribute nilfs_attr_sc_interval = {
	.attr = { .name = "sc_interval", .mode = S_IRUGO | S_IWUSR },
	.show = nilfs_segctor_show_interval,
	.store = nilfs_segctor_store_interval,
};

struct nilfs_attribute nilfs_attr_sc_threshold = {
	.attr = { .name = "sc_threshold", .mode = S_IRUGO | S_IWUSR },
	.show = nilfs_segctor_show_threshold,
	.store = nilfs_segctor_store_threshold,
};

struct nilfs_attribute nilfs_attr_sc_max_bio = {
	.attr = { .name = "sc_max_bio", .mode = S_IRUGO | S_IWUSR },
	.show = nilfs_segctor_show_max_bio,
	.store = nilfs_segctor_store_max_bio,
};

/*
 * Setup & clean-up functions
 */
static struct nilfs_sc_info *
nilfs_segctor_init(struct nilfs_sb_info *sbi, unsigned int max_blocks)
{
	unsigned int bytes;
	struct nilfs_sc_info *sci;

	sci = kmalloc(sizeof(*sci), GFP_KERNEL);
	if (!sci)
		return NULL;

	memset(sci, 0, sizeof(*sci));
	
	bytes = sizeof(struct buffer_head *) * max_blocks;
	sci->sc_bh_arr = kmalloc(bytes, GFP_KERNEL);
	if (!sci->sc_bh_arr) {
		kfree(sci);
		return NULL;
	}

	memset(sci->sc_bh_arr, 0, bytes);

	sci->sc_ba_max = max_blocks;
	sci->sc_sb = sbi->s_super;
	init_completion(&sci->sc_bio_event);

	init_waitqueue_head(&sci->sc_wait_done);
	init_waitqueue_head(&sci->sc_wait_request);
	spin_lock_init(&sci->sc_state_lock);

	sci->sc_interval = HZ * NILFS_SC_DEFAULT_MAX_AGE;
	sci->sc_block_max = NILFS_SC_DEFAULT_THRESHOLD_DATA_AMOUNT;
	sci->sc_flags = NILFS_SEGCTOR_INIT;
	atomic_set(&sci->sc_max_bio, NILFS_SC_DEFAULT_MAX_BIO);

	if (sbi->s_segctor.interval)
		sci->sc_interval = sbi->s_segctor.interval;
	if (sbi->s_segctor.block_max)
		sci->sc_block_max = sbi->s_segctor.block_max;
	return sci;
}

static void nilfs_segctor_destroy(struct nilfs_sc_info *sci)
{
	struct buffer_head **pbh;

	if (!sci)
		return;

	spin_lock(&sci->sc_state_lock);
	if (sci->sc_flags & NILFS_SEGCTOR_INIT)
		goto skip_before_unlock;

	nilfs_kill_segctor_thread(sci);

	if ((sci->sc_flags & (NILFS_SEGCTOR_COMMIT | NILFS_SEGCTOR_KICK)) ||
	    sci->sc_seq_request != sci->sc_seq_done) {
		spin_unlock(&sci->sc_state_lock);
		/* timer is already removed here */
		nilfs_segctor_construct(sci);
		goto skip;
	}
 skip_before_unlock:
	spin_unlock(&sci->sc_state_lock);
 skip:
	if ((pbh = sci->sc_bh_arr) != NULL) {
		while (pbh < sci->sc_bh_arr + sci->sc_ba_max)
			nilfs_brelse(*pbh++);
		kfree(sci->sc_bh_arr);
	}
	kfree(sci);
}

/**
 * nilfs_init_segment - initialize a segment constructor
 * @sbi: nilfs_sb_info
 * @max_blocks: maximum block count of a full segment.
 *
 * nilfs_init_segment() allocates a struct nilfs_sc_info and initilizes it.
 *
 * Return Value: On success, 0 is returned. On error, a following negative
 * error code is returned.
 *
 * %-ENOMEM - Insufficient memory available.
 */
int nilfs_init_segment(struct nilfs_sb_info *sbi, unsigned int max_blocks)
{
	/* Each field of nilfs_segctor is zero-cleared in the initialization of
	   super block info */
	if (!(sbi->s_super->s_flags & MS_RDONLY)) {
		sbi->s_segctor.info = nilfs_segctor_init(sbi, max_blocks);
		if (!sbi->s_segctor.info)
			return -ENOMEM;
	}
	init_rwsem(&sbi->s_segctor.sem);
	spin_lock_init(&sbi->s_segctor.dirty_files_lock);
	INIT_LIST_HEAD(&sbi->s_segctor.dirty_files);
	INIT_LIST_HEAD(&sbi->s_segctor.garbage_list);

	return 0;
}

/**
 * nilfs_clear_segment - destroy a segment constructor
 * @sbi: nilfs_sb_info
 *
 * nilfs_clear_segment() kills the segment constructor daemon,
 * frees the struct nilfs_sc_info, and destroy the dirty file list.
 */
void nilfs_clear_segment(struct nilfs_sb_info *sbi)
{
	/*
	 * Following destruction is not guarded.
	 * Sould we use an exclusion control?
	 */
	nilfs_segctor_destroy(NILFS_SC(sbi));
	sbi->s_segctor.info = NULL;
	
	/* freeing dirty files list */
	down_write(&sbi->s_segctor.sem);
	spin_lock_irq(&sbi->s_segctor.dirty_files_lock);
	if (!list_empty(&sbi->s_segctor.dirty_files)) {
		list_splice(&sbi->s_segctor.dirty_files,
			    &sbi->s_segctor.garbage_list);
		nilfs_warning(sbi->s_super, __FUNCTION__, 
			      "dirty list is not empty\n");
	}
	spin_unlock_irq(&sbi->s_segctor.dirty_files_lock);
	up_write(&sbi->s_segctor.sem);

	dispose_garbage_list(sbi);
}


/* Local Variables:	*/
/* eval: (c-set-style "linux")	*/
/* End:			*/
